/*
 * $Id: qct.cpp,v 1.20 2021/01/18 19:11:56 werdna Exp werdna $
 *
 * Copyright (c) 2016-20, Andrew C Aitchison <qct@aitchison.me.uk>
 *
 */

#pragma clang diagnostic push
// #pragma clang diagnostic ignored "-Winconsistent-missing-destructor-override"
#pragma clang diagnostic ignored "-Wdocumentation-unknown-command"
// #pragma clang diagnostic ignored "-Wweak-vtables"
#pragma clang diagnostic ignored "-Wunused-template"

#include "cpl_port.h"
#include "cpl_error.h"
#include "cpl_string.h"
#pragma clang diagnostic pop

#include "qct.h"

#ifdef CODE_ANALYSIS

// Printing variables with CPLDebug can hide
// the fact that they are not otherwise used ...
#define CPLDebug(...)

#endif // CODE_ANALYSIS


char QCTFGetc(VSILFILE *fp) {
#if 0
    return VSIFGetc(fp);
#else
    char buf;
    size_t res = VSIFReadL(&buf, 1, 1, fp);
#if 0
    CPLDebug("QCT",
             "QCTFGetc(%p): got x%x=%d and res x%zx",
             fp, buf, buf, res);
#endif
    if (1==res) {
        return buf;
    }
    vsi_l_offset nFPos=VSIFTellL(fp);
    CPLError( CE_Failure, CPLE_FileIO,
              "QCTFGetc: failed to read one byte (got %zu) at x%llx=%lld",
              res, nFPos, nFPos);
    //return EOF;
    // Trying to force a crash
    // so that the debugger catches the program before it exits
    // int a = res/res;
    exit(-1);
    // return CE_Failure;
#endif
} // QCTFGetc

unsigned char QCTFGetuc(VSILFILE *fp) {
    return static_cast<unsigned char>(QCTFGetc(fp));
}

#if 0 // Now in qct.h
// Defined in GDAL v1.11.5 but not v1.09 (gdal.system on SL6)
#ifndef CPL_LSBSINT32PTR
#define CPL_LSBSINT32PTR(x) ((GInt32) CPL_LSBINT32PTR(x))
#endif
#ifndef CPL_LSBUINT32PTR
#define CPL_LSBUINT32PTR(x) ((GInt32) CPL_LSBINT32PTR(x))
#endif
#endif // Now in qct.h

extern int QCTGetInt( void* base, int byteOffset )
{
    unsigned char * buf = static_cast<unsigned char*>(base)+byteOffset;
#if 0
    int vv;
    vv = buf[0];
    vv |= (buf[1] << 8u);
    vv |= (buf[2] << 16u);
    vv |= (buf[3] << 24u);

    int lsb=CPL_LSBSINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QCT", "x%x is LSB-endian",vv);
    } else {
        CPLDebug("QCT", "assuming x%x is MSB-endian",vv);
    }

    return(vv);
#else
    return(CPL_LSBSINT32PTR(buf));
#endif
}

unsigned int QCTGetUInt( void* base, int byteOffset )
{
#if 0
    return(static_cast<unsigned int>(QCTGetInt(base, byteOffset )));
#else
    unsigned char * buf = static_cast<unsigned char*>(base)+byteOffset;
    return(CPL_LSBUINT32PTR(buf));
#endif
}

/* -------------------------------------------------------------------------
 * Returns a string (nul-terminated) allocated from VSIMalloc.
 * The string starts at byte byteaddr of the QCT file.
 * If index pointer is nul then a null pointer is returned
 * (rather than an empty string).
 */
extern char *QCTGetString( VSILFILE *fp, size_t byteaddr )
{
    size_t string_length = 0;
    size_t buf_length = 1024;
    if (byteaddr<1) {
        CPLDebug("QCT", "no string at x%zx", byteaddr);
        return nullptr;
    }
    
    char *pszNewString = static_cast<char*>(VSIMalloc(buf_length));
    if (pszNewString == nullptr) {
        CPLError(CE_Failure, CPLE_OutOfMemory,
                 "Cannot allocate space for string");
        exit(-1);
        //return nullptr;
    }
    int seekres = VSIFSeekL( fp,
                             static_cast<vsi_l_offset>(byteaddr),
                             SEEK_SET );
    if ( seekres ) {
        VSIFree(pszNewString);
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QCT string" );
        exit(-1);
        //return nullptr;
    }
    CPLDebug("QCT", "QCTGetString(fp, x%012zx)",byteaddr);
    while (1) {
        size_t bytesread =
            VSIFReadL( &pszNewString[string_length], 1, 1024, fp);
        CPLDebug("QCT",
                 "bytesread %zu string_length %zu buf_length %zu",
                 bytesread, string_length, buf_length);
        while (string_length < buf_length &&
               string_length < bytesread)
            {
                if ( pszNewString[string_length++] == 0 ) {
                    return pszNewString;
                }
            }
        if (bytesread < 1024) {
            VSIFree(pszNewString);
            CPLError(CE_Failure, CPLE_AppDefined,
                     "end of string not found: read %zu/1024 now at %zu/%zu\n",
                     bytesread, string_length, buf_length);
            exit(-1);
            //return nullptr;     
        }
        
        buf_length += 1024;
        pszNewString = static_cast<char*>(VSIRealloc(pszNewString, buf_length));
        if (pszNewString == nullptr) {
            CPLError(CE_Failure, CPLE_OutOfMemory,
                     "Cannot reallocate %zu bytes for string",
                     buf_length);
            exit(-1);
            //return nullptr;
        }
    }

    // return pszNewString;
} // QCTGetString( VSILFILE *fp, size_t byteaddr )


int QCTreadInt(VSILFILE *fp)
{
    // Is this the right-endian ?

    unsigned char buf[4]={};
    size_t ret = VSIFReadL(buf, 1, 4, fp);
    if (ret<4) return(EOF);
    int vv = buf[0];
    vv |= (buf[1] << 8);
    vv |= (buf[2] << 16);
    vv |= (buf[3] << 24);

#if 0 // Noisy debug
    // From cpl_port.h
    int lsb=CPL_LSBSINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QCT", "x%x is LSB-endian", lsb);
    } else {
        CPLDebug("QCT", "assuming x%x is MSB-endian", lsb);
    }
#endif
    return(vv);
}
unsigned int QCTreadUInt(VSILFILE *fp)
{
    // Is this the right-endian ?

    unsigned char buf[4]={};
    size_t ret = VSIFReadL(buf, 1, 4, fp);
    if (ret<4) {
        vsi_l_offset nFPos=VSIFTellL(fp);
        CPLError( CE_Failure, CPLE_FileIO,
                  "QCTreadUInt: failed to read four bytes (got %zu) at x%llx=%lld",
              ret, nFPos-ret, nFPos-ret);
        return CE_Failure;
    }
    unsigned int vv = buf[0];
    vv |= static_cast<unsigned int>(buf[1]) << 8u;
    vv |= static_cast<unsigned int>(buf[2]) << 16u;
    vv |= static_cast<unsigned int>(buf[3]) << 24u;
    // These casts are from uchar to uint;
    // without them the buf[] are promoted to signed int
    
#if 0 // Noisy debug
    // From cpl_port.h
    unsigned int lsb=CPL_LSBUINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QCT", "x%x is LSB-endian", lsb);
    } else {
        CPLDebug("QCT", "assuming x%x is MSB-endian", lsb);
    }
#endif
    return(vv);
}

int QCTreadInt(VSILFILE *fp, vsi_l_offset byteaddr)
{
    int seekres = VSIFSeekL( fp, byteaddr, SEEK_SET );
    if ( seekres ) {
        CPLError( CE_Failure, CPLE_FileIO,
                  "cannot seek to QCT file offset %llu=x%llx: error %d",
                  byteaddr, byteaddr, seekres);
        exit(-1);
        //return 0;
    }
    return QCTreadInt(fp);
}
unsigned int QCTreadUInt(VSILFILE *fp, vsi_l_offset byteaddr)
{
    int seekres = VSIFSeekL( fp, byteaddr, SEEK_SET );
    if ( seekres ) {
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QCT file offset %llu=x%llx: error %d",
                  byteaddr, byteaddr, seekres);
        exit(-1);
        //return 0;
    }
    return QCTreadUInt(fp);
}

signed long long QCTreadInt64(VSILFILE *fp)
{
    // Is this the right-endian ?

    unsigned char buf[8]={};
    size_t ret = VSIFReadL(buf, 1, 8, fp);
    if (ret<8) return(EOF);
    long long vv = buf[0]={};
    vv |= (buf[1] << 8);
    vv |= (buf[2] << 16);
    vv |= (buf[3] << 24);
    vv |= static_cast<long long>(buf[4]) << 32;
    vv |= static_cast<long long>(buf[5]) << 40;
    vv |= static_cast<long long>(buf[6]) << 48;
    vv |= static_cast<long long>(buf[7]) << 56;

#if 0 // Noisy debug
    // From cpl_port.h
    int lsb=CPL_LSBSINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QCT", "x%x is LSB-endian", lsb);
    } else {
        CPLDebug("QCT", "assuming x%x is MSB-endian", lsb);
    }
#endif
    return(vv);
} // QCTreadInt64(VSILFILE *fp)
signed long long QCTreadInt64(VSILFILE *fp, vsi_l_offset byteaddr)
{
    int seekres = VSIFSeekL( fp, byteaddr, SEEK_SET );
    if ( seekres ) {
        CPLError( CE_Failure, CPLE_FileIO,
                  "cannot seek to QCT file offset %llu=x%llx: error %d",
                  byteaddr, byteaddr, seekres);
        exit(-1);
        //return 0;
    }
    return QCTreadInt64(fp);
} // QCTreadInt64(VSILFILE *fp, vsi_l_offset byteaddr)

unsigned long long QCTreadUInt64(VSILFILE *fp)
{
    // Is this the right-endian ?

    unsigned char buf[8]={};
    size_t ret = VSIFReadL(buf, 1, 8, fp);
    if (ret<8) return(EOF);
    // 128 << 24 == -2^31 but 128u << 24 == +2^31 
    unsigned long long vv = static_cast<unsigned long long>(buf[0]);
    vv |= static_cast<unsigned long long>(buf[1]) << 8;
    vv |= static_cast<unsigned long long>(buf[2]) << 16;
    vv |= static_cast<unsigned long long>(buf[3]) << 24;
    vv |= static_cast<unsigned long long>(buf[4]) << 32;
    vv |= static_cast<unsigned long long>(buf[5]) << 40;
    vv |= static_cast<unsigned long long>(buf[6]) << 48;
    vv |= static_cast<unsigned long long>(buf[7]) << 56;

#if 0 // Noisy debug
    // From cpl_port.h
    int lsb=CPL_LSBSINT32PTR(buf);
    if (vv==lsb) {
        CPLDebug("QCT", "x%x is LSB-endian", lsb);
    } else {
        CPLDebug("QCT", "assuming x%x is MSB-endian", lsb);
    }
#endif
    return(vv);
} // QCTreadUInt64(VSILFILE *fp)
unsigned long long QCTreadUInt64(VSILFILE *fp, vsi_l_offset byteaddr)
{
    int seekres = VSIFSeekL( fp, byteaddr, SEEK_SET );
    if ( seekres ) {
        CPLError( CE_Failure, CPLE_FileIO,
                  "cannot seek to QCT file offset %llu=x%llx: error %d",
                  byteaddr, byteaddr, seekres);
        exit(-1);
        //return 0;
    }
    return QCTreadUInt64(fp);
} // QCTreadUInt64(VSILFILE *fp, vsi_l_offset byteaddr)

size_t QCTreadDoubles(double array[],VSILFILE *fp, size_t startaddr, size_t count)
{
    if (array==nullptr || fp==nullptr /* || startaddr<0 || count<0 */) {
        CPLError( CE_Failure, CPLE_IllegalArg,
                  "bad arg in QCTreadDoubles(%p,%p,%zu,%zu)",
                  array, fp, startaddr, count);
        exit(-1);
        //return CE_Failure;
    }

    int seekres = VSIFSeekL( fp, startaddr, SEEK_SET );
    if ( seekres ) {
        int myerr = errno;
        CPLError( CE_Failure, CPLE_AppDefined,
                  "cannot seek to QCT file offset %zu=x%zx: error %d %d",
                  startaddr, startaddr, seekres, myerr);
        return 0;
    }
    size_t entries_read =
        VSIFReadL( array, sizeof(double), count, fp );

    if (entries_read != count) {
        int myerr = errno;
        CPLError( CE_Failure, CPLE_AppDefined,
                  "reading QCT file from offset %zd=x%zx: only read %zd of %zd doubles: error %d",
                  startaddr, startaddr, entries_read, count, myerr);
    }

    // The values are little-endian in the file;
    // convert to endian for this machine.
    for (size_t i=0; i<entries_read; i++) {
        CPL_LSBPTR64(&array[i]);
    }

    return entries_read;
} // QCTreadDoubles()


/* -------------------------------------------------------------------------
 * Convert pixel (x,y) in image into (longitude,latitude)
 * x,y from top left
 * latitude,longitude in degrees WGS84.
 * Does not clip.
 */
void QCT_xy_to_latlon(
                      const double x, double y,
                      latlon *latlon,
                      const double QCTpolynomial[],
                      const double adfQCTdatumShift[]
                      )
{
    double x2 = x * x;
    double x3 = x2 * x;
    double y2 = y * y;
    double y3 = y2 * y;
    double xy = x*y;
    double x2y = x2 * y;
    double xy2 = x * y2;

    // *longitude = lon + lonX * x + lonY * y
    //            + lonXX * x2 + lonXY * x * y + lonYY * y2
    //        + lonXXX * x3 + lonXXY * x2y + lonXYY * xy2 + lonYYY * y3;
    latlon->lon = QCTpolynomial[30] 
        + QCTpolynomial[31] * x + QCTpolynomial[32] * y
        + QCTpolynomial[33] * x2 + QCTpolynomial[34] * xy
            + QCTpolynomial[35] * y2
        + QCTpolynomial[36] * x3 + QCTpolynomial[37] * x2y
            + QCTpolynomial[38] * xy2 + QCTpolynomial[39] * y3;
    // *latitude = lat + latX * x + latY * y
    //           + latXX * x2 + latXY * x * y + latYY * y2 
    //         + latXXX * x3 + latXXY * x2y + latXYY * xy2 + latYYY * y3;
    latlon->lat = QCTpolynomial[20] 
        + QCTpolynomial[21] * x  + QCTpolynomial[22] * y
        + QCTpolynomial[23] * x2 + QCTpolynomial[24] * xy
            + QCTpolynomial[25] * y2
        + QCTpolynomial[26] * x3 + QCTpolynomial[27] * x2y
            + QCTpolynomial[28] * xy2 + QCTpolynomial[29] * y3;
    
    // Add the datum shift
    latlon->lon += adfQCTdatumShift[1];
    latlon->lat += adfQCTdatumShift[0];
} // QCT_xy_to_latlon()



QC3Header::QC3Header() :
    // osQC3Filename(nullptr),
    fpQC3(nullptr),
    magic3(0),
    version3(0),
    width3(0),
    height3(0),
    encrypt3(-1),
    reserved{0,0,0,0,0},
    nQC3filesize(0)
{
} // QC3Header::QC3Header

QC3Header::~QC3Header()
{
    if (fpQC3 != nullptr) {
        VSIFCloseL( fpQC3 );
        fpQC3 = nullptr;
    }
}

int QC3Header::Open( CPLString osQCTFilename,
                      int bUpdate, int nQCTmagic, int nQCTVersion )
{
    if (bUpdate) {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                  "Update access not supported by the QCT driver." );
        return false;
    }
    CPLDebug("QCT", "QC3Header::Open(%s %d x%x x%x)",
             osQCTFilename.c_str(), bUpdate, nQCTmagic, nQCTVersion);
    CPLString osPathname = CPLGetPath( osQCTFilename.c_str() );
    CPLString osBasename = CPLGetBasename( osQCTFilename );
    osQC3Filename = CPLFormFilename( osPathname, osBasename, "qc3" );
    CPLDebug("QCT", "nQCTVersion x%08x %s",
             nQCTVersion, osQC3Filename.c_str());

    // width3=0;     height3=0;
    // fpQC3 = nullptr;
    // nQC3filesize=0;
    VSIStatBufL osRscStatBuf3={};
    if ( VSIStatL( osQC3Filename.c_str(), &osRscStatBuf3 ) != 0 ) {
        CPLDebug("QCT", "No QC3 file %s %d",
                 osQC3Filename.c_str(), errno);
        return false;
    }
    if (osRscStatBuf3.st_size<40) {
        CPLDebug("QCT", "QC3 file %s exists but is too small.",
                 osQC3Filename.c_str());
        return false;
    }
    if ( (nQCTVersion & qct_version_for_qc3_mask) != qct_version_for_qc3 ) {
        CPLDebug("QCT", "qct file version x%08x - expected x%08x",
                 nQCTVersion, qct_version_for_qc3);
        // return false;
    }
    if ( nQCTmagic != qct_info_magic && nQCTmagic != qct_magic ) {
        CPLDebug("QCT", "qct file magic x%08x - expected x%08x",
                 nQCTmagic, qct_magic);
        // return false;
    }

    fpQC3 = VSIFOpenL( osQC3Filename.c_str(), "r" );
    if (fpQC3 == nullptr ) {
        CPLDebug("QCT", "%s failed to open",
                 osQC3Filename.c_str());
        return false;
    }

    nQC3filesize=osRscStatBuf3.st_size;
    CPLDebug("QCT", "QC3 file %s opened successfully",
             osQC3Filename.c_str() );
    unsigned int magic3 = QCTreadUInt(fpQC3);
    unsigned int version3 = QCTreadUInt(fpQC3);
    if (magic3 != qcthree_magic || version3 != 1) {
        CPLDebug("QCT",
                 "QC3: expected magic x%08x version x%08x, but read x%08x x%08x",
                 qcthree_magic, 1, magic3, version3);
        return false;
    }

    encrypt3 = QCTreadInt(fpQC3);
    if (encrypt3>=10) {
        CPLDebug("QCT",
                 "QC3 only has 10 overviews; data cannot be encrypted from overview %d",
                 encrypt3);
        encrypt3=9;
    }
    if (encrypt3>=0) {
        CPLDebug("QCT",
             "QC3 data encrypted below overview %d",
             encrypt3+1);
    } else {
        CPLDebug("QCT", "QC3 data not encrypted");
    }

    width3 = QCTreadUInt(fpQC3, 12);
    height3 = QCTreadUInt(fpQC3);
    // Could read next 5 ints and verify they are null

    const int tenmillion = 10*1000*1000;
    if (width3 > 4*tenmillion || height3 > 2*tenmillion) {
        // Arbitrary cut-off
        // Enough to cover the world with one 1024x1024 tile
        // per square metre !
        CPLDebug("QCT",
                 "implausible QC3 file size %u x %u 1024 pixel tiles ignored",
                 width3, height3
                 );
        width3=0;     height3=0;
        // QCT file still OK            
    } else {
        CPLDebug("QCT",
                 "QC3 file is %u x %u 1024 pixel tiles",
                 width3, height3
                 );
    }
    
    return true;
} // QC3Header::Open()

/* These are, possibly, the coordinate reference systems of the *maps*.
 * All qct/3 files are specified in EPSG:4236 - not true
 * Datum Projection
 * Metadata Flags Depths Heights Keywords
 */
const char* QCT_copyright_to_crs[] =
    {
     // We can ignore trailing strings (eg ".") as we will use a substring match.
     // UK
     "2002 Civil Aviation Authority. Crown Copyright; Ordnance Survey", "EPSG:27700",
     "2003 Geographers' A-Z Map Company Limited", "EPSG:27700",
     "2006 Geographers A-Z Map Company Limited", "EPSG:27700",
     "2006 Crown Copyright; Ordnance Survey, Licence Number PU 100034184", "EPSG:27700",
     "2012 Crown Copyright; UK Hydrographic Office, HMSO, Memory-Map Europe. Not to be used for navigation", "",
     "Cassini Publishing Ltd and Memory-Map Europe", "EPSG:27700",
     "UNITED KINGDOM HYDROGRAPHIC OFFICE/MAPTECH INC", "",
     "Waterway Routes.  Contains Ordnance Survey data.  Crown copyright and database right 2013", "EPSG:27700",
     // Photomaps, UK
     "2002 Getmapping plc & EVO Distribution Ltd", "EPSG:27700",
     // Australia
     // MGA94
     // GDA94 EPSG:3577 EPSG:4283
     // AMG66
     // AGD66 EPSG:4202 - zones 48-58 EPSG:20248 - EPSG:20258
     // EPSG:3577 EPSG:4283 EPSG:4202
     "South Australia Government (DEWNR) 2016","EPSG:3577",
     "State of Queensland (Department of Natural Resources, Mines and Energy) 2018","EPSG:3577",
     "Crown Copyright, NSW Department Finance, Services and Innovation, 2018","EPSG:3577",
     "2019 State of Victoria","+proj=lcc +lat_0=-37 +lon_0=145 +x_0=2500000 +y_0=2500000 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_def", // My guess at VicGrid94 with tmerc, which is not quite EPSG:3111
     "State of Tasmania,Base Data and Raster images from the LIST (www.thelist.tas.gov.au)","EPSG:3577",
     "2012 Commonwealth of Australia, Copyright restrictions apply. WARNING. The Commonwealth of Australia accepts no liability for the accuracy of information contained in this product and does not warrant that this product meets any regulations as an appropriate product for navigation. For full details, refer to the documentation and packaging accompanying this product.","EPSG:3577",
     // New Zealand
     // Try 4167 4959 4272 3851 27200 2958
     // +proj=nzmg +lat_0=0 +lon_0=0 +x_0=0 +y_0=0 +ellps=intl +towgs84=59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993 +units=m +no_defs
     "Crown Copyright Reserved, Land Information New Zealand",
     "EPSG:4958", /* "EPSG:4364",  "EPSG:27200", */
     "2004, Terralink", "",
     // Belgium
     "NGI Belgium Reproduction or adaptation in any form is forbidden. Memory-Map Europe", "",
     // France
     "IGN 2006 - Reproduction and/or sale prohibited", "EPSG:2154",
     // Norway
     "Kartverket", "EPSG:25833",
     // Spain
     "2006 David Brawn", "",
     "2009 David Brawn", "",
     "2009 David Brown", "",
     "2010 David Brown", "",
     "2011 David Brawn", "",
     "Copyright Discovery Walking Guides Ltd", "",
     // Europe
     "2003 Philips.", "",
     // USA, World
     //  try EPSG:5070/5071/5072 for USA maps
     "2004, Memory-Map, Inc", "",
     // South Africa, Mexico
     "GoTrekkers", "",
     // 
     "","",
    }; // QCT_copyright_to_crs

extern int EPSG_from_Copyright(char* pszCopyright)
{
    const char ** ppszLoop = &QCT_copyright_to_crs[0];
    while (*ppszLoop) {
        if (strstr(*ppszLoop, pszCopyright) && ppszLoop[1]) {
            const char* pEPSGStart = strstr(ppszLoop[1], "EPSG:");
            if (pEPSGStart && pEPSGStart[5]) {
                return atoi(pEPSGStart+5);
            }
        }
        ppszLoop+=2;
    }
    return 0;
}
