/*
 * $Id: qct.h,v 1.14 2021/01/18 19:09:31 werdna Exp werdna $
 *
 * Copyright (c) 2016-20, Andrew C Aitchison <qct@aitchison.me.uk>
 *
 */


#ifndef QCT_H_INCLUDED
#define QCT_H_INCLUDED

#ifndef QCT_USE_GCP
#define QCT_USE_GCP 1
#endif

#define QCT_TILE_SIZE 64
#define QCT_TILE_PIXELS (QCT_TILE_SIZE*QCT_TILE_SIZE)
#define QC3_TILE_SIZE 1024
#define QC3_TILE_PIXELS (QC3_TILE_SIZE*QC3_TILE_SIZE)

#define qct_info_magic 0x1423d5fe
#define qct_magic      0x1423d5ff // 0xffd52314
#define qcthree_magic  0x484df282
#define qed_magic      0x34eb3498
#define qct_version_for_qc3_mask 0xfffffffe
#define qct_version_for_qc3 0x20000000
// The usual value of qct_version_for_qc3 is 0x20000001 but it can be 0x20000000
#include <cpl_vsi.h>

typedef struct { double lat; double lon; } latlon;

#if GDAL_VERSION_MAJOR < 2
typedef  long fid_t;
#define  fid_f "%ld"
#else
typedef  GIntBig fid_t;
#define  fid_f "%lld"
#endif

// Defined in GDAL v1.11.5  but not v1.09
#ifndef CPL_LSBSINT32PTR
#define CPL_LSBSINT32PTR(x) ((GInt32) CPL_LSBINT32PTR(x))
#endif
#ifndef CPL_LSBUINT32PTR
#define CPL_LSBUINT32PTR(x) ((GInt32) CPL_LSBINT32PTR(x))
#endif

char QCTFGetc(VSILFILE *fp);
unsigned char QCTFGetuc(VSILFILE *fp);

extern int QCTGetInt( void* base, int byteOffset );
unsigned int QCTGetUInt( void* base, int byteOffset );

int QCTreadInt(VSILFILE *fp);
int QCTreadInt(VSILFILE *fp, vsi_l_offset byteaddr);
unsigned int QCTreadUInt(VSILFILE *fp);
unsigned int QCTreadUInt(VSILFILE *fp, vsi_l_offset byteaddr);

long long QCTreadInt64(VSILFILE *fp);
long long QCTreadInt64(VSILFILE *fp, vsi_l_offset byteaddr);
unsigned long long QCTreadUInt64(VSILFILE *fp);
unsigned long long QCTreadUInt64(VSILFILE *fp, vsi_l_offset byteaddr);

size_t QCTreadDoubles(double array[], VSILFILE *fp, size_t startaddr, size_t count);
extern char *QCTGetString(VSILFILE *fp, size_t byteaddr);

void 
QCT_xy_to_latlon(
                 const double x, const double y,
                 latlon *latlon,
                 const double QCTpolynomial[],
                 const double adfQCTdatumShift[]
                 );
#if 0
void 
QCT_latlon_to_xy(
                 double x, double y,
                 const latlon const *latlon,
                 const double QCTpolynomial[],
                 const double adfQCTdatumShift[]
                 );
#endif


// extern
class QC3Header
{
    // friend class QCTDataSource;
    friend class QCTDataset;
    // friend class OGRQCTDataSource;
    friend class OGRQCTLayer;
    friend class QCTRasterBand;
    CPLString       osQC3Filename;
    VSILFILE       *fpQC3;
    unsigned int    magic3;
    unsigned int    version3;
    unsigned int    width3;
    unsigned int    height3;
      signed int    encrypt3;
    unsigned int    reserved[5]; // set to 0
    vsi_l_offset    nQC3filesize;

public:
    QC3Header();
    ~QC3Header();
    int Open(const CPLString osQCTFilename,
             int bUpdate,
             int nQCTMagic,
             int nQCTVersion
             );
}; // class QC3Header

extern int EPSG_from_Copyright(char* pszCopyright);

#endif // ifndef QCT_H_INCLUDED
